/*
 *
 * OpenSBC.cxx
 *
 * Open SIP Stack ( OSS )
 *
 * Copyright (c) opensipstack.org.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is OpenSIPStack.
 *
 * The Initial Developer of the Original Code is opensipstack.org.
 *
 * The author of this code is Joegen E. Baclor
 *
 * $Log: OpenSBC.cxx,v $
 * Revision 1.64  2007/07/30 06:47:29  joegenbaclor
 * removed unused variable uproute
 *
 * Revision 1.63  2007/07/29 06:29:55  joegenbaclor
 * updated visual studio 2003 project files
 *
 * Revision 1.62  2007/07/25 12:20:31  joegenbaclor
 * changed conf file version to 1.1.4-2
 *
 * Revision 1.61  2007/07/20 03:33:14  joegenbaclor
 * Config aesthetics
 *
 * Revision 1.60  2007/07/17 02:01:22  joegenbaclor
 * modified config template
 *
 * Revision 1.59  2007/07/17 01:31:48  joegenbaclor
 * Added Config version checking
 *
 * Revision 1.58  2007/07/15 12:27:25  joegenbaclor
 * Added capability to route to registered endpoints using static routes
 * Removed unused config parameters
 *
 * Revision 1.57  2007/07/15 07:51:20  joegenbaclor
 * implemented resource counter page
 *
 * Revision 1.56  2007/07/14 07:53:56  joegenbaclor
 * Migrated code to use new STL based string class
 *
 * Revision 1.55  2007/07/06 14:18:25  rcolobong
 * Add Upstream Proxy Route feature
 *
 * Revision 1.54  2007/06/17 02:49:18  joegenbaclor
 * Introduced Global lock to PHTTPConfig
 *
 * Revision 1.53  2007/06/16 13:06:19  joegenbaclor
 * change regdb to use CachePointers instead fo a direct pointer access to cache objects
 *
 * Revision 1.52  2007/06/13 09:22:01  joegenbaclor
 * Added IVR handler
 *
 * Revision 1.51  2007/06/06 16:20:31  joegenbaclor
 * Commiting Media Server Related Changes
 *
 * Revision 1.50  2007/05/18 01:23:44  joegenbaclor
 * 1.  Added separate useragent to handle B2BUpperReg backdoor
 * 2.  Modified routing for B2BUpperReg
 * 3.  Corrected bug in ParserTools::GetCompactHeader()
 *
 * Revision 1.49  2007/05/07 06:13:22  rcolobong
 * 1. Add support for disabling and enabling Local Refer
 * 2. Add support for defining static RTP media address
 * 3. Remove support for Force B2B Routes
 *
 * Revision 1.48  2007/05/06 15:53:16  joegenbaclor
 * Incremented minor version
 *
 * Revision 1.47  2007/05/04 10:07:18  joegenbaclor
 * more on delayed deletion mechanism
 *
 * Revision 1.46  2007/04/11 09:11:38  joegenbaclor
 * Corrected type from last commit
 *
 * Revision 1.45  2007/04/10 06:31:46  joegenbaclor
 * Added new Encryption Mode
 *
 * Revision 1.44  2007/03/08 10:27:09  joegenbaclor
 * Tidy up OpalManager Destruction
 * Enabled Recording in vxml
 *
 * Revision 1.43  2007/03/08 06:43:20  joegenbaclor
 * This commit marks first successful call to the Media Server
 *
 * Revision 1.42  2007/03/07 23:20:13  joegenbaclor
 * Added media server UA
 *
 * Revision 1.41  2007/02/22 14:41:13  joegenbaclor
 * Completed multi listener support
 *
 * Revision 1.40  2007/02/22 08:23:51  joegenbaclor
 * Added multi listener support
 *
 * Revision 1.39  2007/02/21 01:00:02  joegenbaclor
 * Fixed compile error in CentOS.  Thanks Harold Bray
 *
 * Revision 1.38  2007/02/16 11:09:19  joegenbaclor
 * More work on privacy
 *
 * Revision 1.37  2007/02/15 09:36:29  joegenbaclor
 * Added Privacy Section
 *
 * Revision 1.36  2007/02/15 02:51:44  joegenbaclor
 * Completed trusted domain and proxy authentication support
 *
 * Revision 1.35  2007/02/14 11:22:08  joegenbaclor
 * Added Trusted Domain section
 *
 * Revision 1.34  2007/02/08 18:12:21  joegenbaclor
 * Appropriately rename connection timeout to SeizeTimeout
 *
 * Revision 1.33  2007/01/29 08:37:02  joegenbaclor
 * Added TCP Listener activation code
 *
 * Revision 1.32  2007/01/23 11:10:59  joegenbaclor
 * Last minute changes for new tarball release
 *
 * Revision 1.31  2007/01/22 10:00:58  joegenbaclor
 * Fixed ProxyRouteRequest
 *
 * Revision 1.30  2007/01/18 08:39:27  joegenbaclor
 * Added VC7.10 project files for launcher
 *
 * Revision 1.29  2007/01/18 06:06:41  joegenbaclor
 * Incremented Build Number for latest development release
 *
 * Revision 1.28  2007/01/17 10:33:43  joegenbaclor
 * Redesigned syslog
 *
 * Revision 1.27  2007/01/17 00:04:31  joegenbaclor
 * Added SysLog
 *
 * Revision 1.26  2007/01/15 10:06:57  joegenbaclor
 * More on XML-RPC Registration Queries
 *
 * Revision 1.25  2007/01/15 09:24:47  joegenbaclor
 * Added capability to find Registraton via XML-RPC
 *
 * Revision 1.24  2007/01/14 13:15:09  joegenbaclor
 * Fixed bug were mode is not initialized properly for configuration file
 *
 * Revision 1.23  2007/01/12 10:50:04  joegenbaclor
 * minor proxy bug fixes
 * more launcher code
 *
 * Revision 1.22  2007/01/12 01:02:44  joegenbaclor
 * launcher related code
 *
 * Revision 1.21  2007/01/11 10:00:26  joegenbaclor
 * more launcher code
 *
 * Revision 1.20  2007/01/11 02:34:58  rcolobong
 * Add new http config for enable/disable proxying media
 *
 * Revision 1.19  2007/01/10 23:58:23  joegenbaclor
 * launcher specific code
 *
 * Revision 1.18  2007/01/10 05:00:39  joegenbaclor
 * Added B2BUpperReg mode.
 *
 * Revision 1.17  2007/01/09 02:29:44  joegenbaclor
 * Fixed Unregister Post code
 *
 * Revision 1.16  2007/01/08 11:15:18  joegenbaclor
 * Added registration status page
 *
 * Revision 1.15  2007/01/08 07:13:23  joegenbaclor
 * Added ability to run SBC in pure proxy or pure b2b modes
 *
 * Revision 1.14  2007/01/02 01:54:48  joegenbaclor
 * Corrected parser bug for Route Records
 *
 * Revision 1.13  2006/12/22 12:05:42  joegenbaclor
 * We need to include all params in arg.parse in daemon constructor or unix implementation would break.  It seems Parse of PargList can only be called one.  Second call the Parse wont update the params.
 *
 * Revision 1.12  2006/12/22 07:40:27  joegenbaclor
 * More on command line argument parsing.
 *
 * Revision 1.11  2006/12/21 08:56:59  joegenbaclor
 * Added pid in b2bua log file name
 * Fixed bug in proxy relay where startline URI is not rewritten if a static route is
 *  not found
 *
 * Revision 1.10  2006/11/25 02:59:13  joegenbaclor
 * Added project files to opensbc.  Old scheme of producing the VC 8.0 workspace using configure is still available
 *
 * Revision 1.9  2006/11/24 10:59:55  rcolobong
 * Remove support for RTBE Classes
 *
 * Revision 1.8  2006/11/22 11:37:09  rcolobong
 * 1. Added new support for Force B2BUA in HTTP configuration
 * 2. Added log level in Application level and SIP/RTP Level
 * 3. Update OnAcceptRegistration signature
 *
 * Revision 1.7  2006/10/11 04:59:50  rcolobong
 * Add http configuration "Accept all registration". If this true then it will accept all registration even if it is not included on the "Local domain accounts"
 *
 * Revision 1.6  2006/09/04 14:39:07  joegenbaclor
 * Corrected bug where logger prefix defaults to b2bua--
 *
 * Revision 1.5  2006/09/01 12:51:03  rcolobong
 * Use ActiveSessionCounter to update session count through OSSApplication
 *
 * Revision 1.4  2006/08/30 09:07:31  rcolobong
 * Update setting connection timer and alerting timer
 *
 * Revision 1.3  2006/08/29 07:43:58  rcolobong
 * 1. Rename file to a better filename
 * 2. Change SBCRTBESesionManager class to RTBESessionManager
 *
 * Revision 1.2  2006/08/16 10:34:32  rcolobong
 * 1. Syslog server is now configurable
 * 2. Alerting timer and Connection Timer is now configurable
 *
 * Revision 1.1  2006/08/14 10:04:59  rcolobong
 * Convert B2BUA to SBC
 * Support MP logging
 *
 * Revision 1.10  2006/08/04 07:30:52  joegenbaclor
 * #ifdefed RTBE headers
 *
 * Revision 1.9  2006/08/04 05:44:18  rcolobong
 * Instantiate B2BRTBEUA instead of openB2BUA if HAS_RTBECLIENT is defined
 *
 * Revision 1.8  2006/07/26 09:19:41  joegenbaclor
 * Added System Log support
 *
 * Revision 1.7  2006/07/17 11:36:44  joegenbaclor
 * 1.  More routing enhancements to B2BUA
 *
 * Revision 1.6  2006/07/11 13:59:34  joegenbaclor
 * Completed Proxy functionalities for B2BUA
 *
 * Revision 1.5  2006/07/10 06:29:37  joegenbaclor
 * Completed crude Registration support for B2BUA
 *
 * Revision 1.4  2006/07/03 15:29:47  joegenbaclor
 * Started merging OpenB2BUA into OpenSBC
 *
 * Revision 1.3  2006/06/29 00:50:25  joegenbaclor
 * Corrected some typos that makes unix compilation barf
 *
 * Revision 1.2  2006/06/20 09:58:11  joegenbaclor
 * *** empty log message ***
 *
 *
 */


#include "OpenSBC.h"
#include "SBCRoutingHandler.h"
#include "SBCCallHandler.h"
#include "SBCAuthHandler.h"
#include "SBCIVRHandler.h"
#include "Version.h"
#include "ossbuildopts.h"
#include "Encryption.h"
#include "ResourceCounter.h"


#define new PNEW

#define XML_CONFIG_ID "1.1.4-2"

static const char configKeySection[] = "OpenSBC General Parameters";
static const char configKeyInterfaceAddress[] = "Interface Address";
static const char configKeyTransactionThreadCount[] = "Transaction Thread Count";
static const char configKeySessionThreadCount[] = "Session Thread Count";
static const char configKeyAppLogLevel[] = "SIP Log Level";
static const char configKeyCoreLogLevel[] = "PTRACE Log Level";
static const char configKeySysLogServer[] = "Syslog Server";
static const char configKeyStaticRTPMediaAddress[] = "Static RTP Media Address";
static const char configKeyEnableLocalRefer[] = "Enable Local Refer";
static const char configKeySeizeTimeout[] = "Seize Timeout";
static const char configKeyAlertingTimeout[] = "Alerting Timeout";
static const char configKeyAppLogFilePrefix[] = "Log File Prefix";
static const char configKeySIPSessionTimer[] = "SIP Session Timer";


static const char configKeyEncryption[] = "Encryption Mode";
static const char configKeyEncryptionKey[] = "Encryption Key";

static const char configKeySBCMode[] = "SBC Mode";
static const char configKeyMediaProxy[] = "Always Proxy Media";

static const char configKeyAccountSection[] = "Local Domain Accounts";
static const char configKeyAcceptAllRegistration[] = "Accept All Registration";
static const char configKeyAccount[] = "Account";

static const char configKeyTrustedDomainSection[] = "Trusted Domains";
static const char configKeyAcceptAllCalls[] = "Accept All Calls";
static const char configKeyDomain[] = "Domain";

static const char configKeyPrivacyTrustedDomainSection[] = "Privacy (RFC 3325)";
static const char configKeyPrivacyDefaultRealm[] = "Default Realm";
static const char configKeyPrivacyDomain[] = "Domain";

static const char configKeyB2BUASection[] = "Force B2BUA Routes";
static const char configKeyB2BUARoutes[] = "Route";

static const char configAutoAttendantSection[] = "Media Server";
static const char configKeyVXMLScript[] = "VXML Script";

OpenSBCDaemon::OpenSBCDaemon()
{
  m_IsInitialized = FALSE;
  m_HasExtendedArgs = FALSE;
  m_SBC = NULL;
  m_EndPoint = NULL;
  m_BackDoorEndPoint = NULL;
  m_Registrar = NULL;
  m_Proxy = NULL;
  m_MediaServer = NULL;

  PArgList & args = GetArguments();
  
  args.Parse("a-app-name:"
             "v-version."
             "d-daemon."
             "c-console."
             "h-help."
             "x-execute."
             "p-pid-file:"
             "H-handlemax:"
             "i-ini-file:"
             "k-kill."
             "t-terminate."
             "s-status."
             "l-log-file:"
             "u-uid:"
             "g-gid:"
             "C-core-size:"
             "D-debug."
             "P-http-port:"
             "X-misc-param-1:"
             "Y-misc-param-2:"
             "Z-misc-param-3:");

  PStringStream prodName;
  prodName << "OpenSBC";

  if( args.HasOption( 'P' ) )
    prodName << " [Port:" << args.GetOptionString( 'P' ) << "]";

  SetProductName( prodName );
  SetMajorVersion( SBC_MAJOR_VERSION );
  SetMinorVersion( SBC_MINOR_VERSION );
  SetBuildNumber( SBC_BUILD_NUMBER );

#ifdef WIN32
  if (args.HasOption('l')) 
  {
    systemLogFileName = args.GetOptionString('l');
    if (systemLogFileName.IsEmpty()) 
    {
      cout << "error: must specify file name for -l" << endl;
    }
    else if (PDirectory::Exists(systemLogFileName))
      systemLogFileName = PDirectory(systemLogFileName) + PProcess::Current().GetFile().GetFileName() + ".log";
  }else
  {
    systemLogFileName = GetFile().GetDirectory() + "opensbc.log";
  }
#endif

  if( args.HasOption('a') )
  {
    OString appName = args.GetOptionString('a');
    
    PStringStream stream;
    stream <<  PProcess::GetName() << " (" <<  appName << ")";
    m_ExtendedAppName = stream;
  }
  
}

OpenSBCDaemon::~OpenSBCDaemon()
{
  if( m_SBC != NULL )
    delete m_SBC;

  if( m_EndPoint != NULL )
    delete m_EndPoint;

  if( m_BackDoorEndPoint != NULL )
    delete m_BackDoorEndPoint;


  if( m_Registrar != NULL )
    delete m_Registrar;

  if( m_Proxy != NULL )
    delete m_Proxy;

  if( m_MediaServer != NULL )
    delete m_MediaServer;
}

// pure virtual from OSSApplication
void OpenSBCDaemon::OnStart( 
  OSSAppConfig & config 
)
{
  OString confVersionId = config.GetVersionId();
  if( confVersionId.IsEmpty() )
    confVersionId = "(empty)";
  
  if( config.GetVersionId() != XML_CONFIG_ID )
  {
    PTRACE( 1, "*** FATAL ERROR *** Attempt to load incorrect version of oss-application.conf.xml!!!" );
    PTRACE( 1, "*** FATAL ERROR *** oss-application.conf.xml version " << confVersionId << " FOUND." );
    PTRACE( 1, "*** FATAL ERROR *** oss-application.conf.xml version " << XML_CONFIG_ID << " EXPECTED!!!" );
    exit( 1 );
  }
  
  if( HasExtendedArgs() )  
    m_InstanceId = m_ExtendedArgs.GetParameter( "iid" );

  int logLevel = config.GetInteger( configKeySection, configKeyAppLogLevel, 3 );
  PTRACE( 1, "Setting SIP Log Level to " << logLevel );
  
  OString logPrefix = GetLogDirectory() + "/" + config.GetString( 
    configKeySection, configKeyAppLogFilePrefix, "b2bua" ) ;

  if( !m_InstanceId.IsEmpty() )
    logPrefix = (const char *)(GetLogDirectory() + "/" + m_InstanceId);
  
  if( m_SBC == NULL )
  {
    OString mode = m_ExtendedArgs.GetParameter( "ua-mode" );
  
    if( mode.IsEmpty() )
      mode = (const char *)config.GetString( configKeySection, configKeySBCMode, "Full Mode" );

    B2BUserAgent::UAMode uaMode = B2BUserAgent::FullMode;

    if( mode == "Full Mode" || mode == "FullMode" )
      uaMode = B2BUserAgent::FullMode;
    else if( mode == "Proxy Only Mode" || mode == "ProxyOnlyMode" )
      uaMode = B2BUserAgent::ProxyOnlyMode;
    else if( mode == "B2B Only Mode" || mode == "B2BOnlyMode" )
      uaMode = B2BUserAgent::B2BOnlyMode;
    else if( mode == "B2BUpperReg Mode" || mode == "B2BUpperRegMode" )
      uaMode = B2BUserAgent::B2BUpperRegMode;

    if( ( uaMode == B2BUserAgent::ProxyOnlyMode || 
          uaMode == B2BUserAgent::B2BOnlyMode ) &&
          config.GetType() == OSSAppConfig::XMLRPCClientProvider )
    {
      OString addr = m_ExtendedArgs.GetParameter( "syslog" );
      if( addr.IsEmpty() )
        addr = "127.0.0.1:514";
      
      SIPURI sysLogAddress( addr );
      Logger::SetDefaultLogStream( 
        new LoggingSystemLogStream( 
          m_InstanceId, 
          PIPSocket::Address( "127.0.0.1"  ),
          sysLogAddress.GetAddress(),
          (WORD)sysLogAddress.GetPort().AsInteger(),
          TRUE ) );

    }else
    {
      Logger::SetDefaultLogStream( new LoggingIncrementingFileStream( logPrefix.c_str(), OString( GetProcessID() ).c_str() ) );
    }

    m_SBC = new OpenSBC( this, uaMode );

    if( uaMode == B2BUserAgent::FullMode )
    {
      PTRACE( 1, "Running in Full Mode" );
    }else if( uaMode == B2BUserAgent::ProxyOnlyMode )
    {
      PTRACE( 1, "Running in Proxy Mode" );
    }else if( uaMode == B2BUserAgent::B2BOnlyMode )
    {
      PTRACE( 1, "Running in Back To Back Mode" );
    }else if( uaMode == B2BUserAgent::B2BUpperRegMode )
    {
      PTRACE( 1, "Running in Back To Back with Upper Registration Mode" );
    }

    if( m_EndPoint == NULL )
      m_EndPoint = new B2BUAEndPoint( *m_SBC, config.GetInteger( configKeySection, configKeySessionThreadCount, 10 ), 1024 * 2 );

    m_SBC->SetEndPoint( m_EndPoint );

    if( m_BackDoorEndPoint == NULL )
      m_BackDoorEndPoint = new B2BUAEndPoint( *m_SBC, config.GetInteger( configKeySection, configKeySessionThreadCount, 10 ), 1024 * 2 );
    
    m_BackDoorEndPoint->SetUACoreName( "BackDoor" );
    m_SBC->SetBackDoorEndPoint( m_BackDoorEndPoint );

    if( m_Registrar == NULL )
      m_Registrar = new B2BUserAgent::Registrar( *m_SBC, config.GetInteger( configKeySection, configKeySessionThreadCount, 10 ), 1024 * 2 );

    m_SBC->SetRegistrar( m_Registrar );

    if( m_Proxy == NULL )
      m_Proxy = new B2BUserAgent::Proxy( *m_SBC, config.GetInteger( configKeySection, configKeySessionThreadCount, 10 ), 1024 * 2 );

    if( m_MediaServer == NULL )
    {
      m_MediaServer = new MS::MediaServer( m_EndPoint );
      m_SBC->SetMediaServer( m_MediaServer );
    }

    m_SBC->SetProxy( m_Proxy );

    m_SBC->OnStart( config );

  }


  PTRACE( 1, "OpenSBC STARTED " << m_InstanceId ); 
}

void OpenSBCDaemon::OnStop()
{
  m_SBC->OnStop();
  
  if( m_SBC != NULL )
  {
    m_SBC = NULL;
    delete m_SBC;
  }

  if( m_EndPoint != NULL )
  {
    m_EndPoint->GetUserAgent().Terminate();
    delete m_EndPoint;
    m_EndPoint = NULL;
  }

  if( m_BackDoorEndPoint != NULL )
  {
    m_BackDoorEndPoint->GetUserAgent().Terminate();
    delete m_BackDoorEndPoint;
    m_BackDoorEndPoint = NULL;
  }

  if( m_Registrar != NULL )
  {
    delete m_Registrar;
    m_Registrar = NULL;
  }

  if( m_Proxy != NULL )
  {
    delete m_Proxy;
    m_Proxy = NULL;
  }

  if( m_MediaServer != NULL )
  {
    delete m_MediaServer;
    m_MediaServer = NULL;
  }



  OSSApplication::OnStop();
}

OSSAppConfig * OpenSBCDaemon::OnCreateAppConfig()
{
  OString configMode = m_ExtendedArgs.GetParameter( "config-mode" );
  OString xmlRPCPort =  m_ExtendedArgs.GetParameter( "xml-rpc-port" );
  OString xmlRPCServer = m_ExtendedArgs.GetParameter( "xml-rpc-server" );
  OString instanceId = m_ExtendedArgs.GetParameter( "iid" );

  WORD port = static_cast<WORD>(xmlRPCPort.IsEmpty() ? 10000 : xmlRPCPort.AsInteger());
  if( configMode.IsEmpty() )
  {
    return new OSSAppConfig( OSSAppConfig::HTTPDefaultProvider, port );
  }else
  {
    if( (configMode *= "XMLRPCClientProvider") && !xmlRPCServer.IsEmpty() )
    {
      return new OSSAppConfig( OSSAppConfig::XMLRPCClientProvider, port, xmlRPCServer, instanceId );
    }else if( configMode *= "XMLRPCAndHTTPServerProvider" )
    {
      return new OSSAppConfig( OSSAppConfig::XMLRPCAndHTTPServerProvider, port );
    }else if( configMode *= "XMLRPCServerProviderOnly" )
    {
      return new OSSAppConfig( OSSAppConfig::XMLRPCServerProviderOnly, port );
    }
  }

  return new OSSAppConfig( OSSAppConfig::HTTPDefaultProvider, port );
}

#if defined(_MSC_VER)
#pragma warning(push)
#pragma warning(disable:4311)
#endif

void OpenSBCDaemon::OnConfigChanged( 
  OSSAppConfig & config 
)
{
  PThread::Create( 
      PCREATE_NOTIFIER( 
      OnConfigChanged ), 
      (INT)(OSSAppConfig*)&config, 
      PThread::AutoDeleteThread,
      PThread::NormalPriority,
      "OnConfigChanged"  );

}

#if defined(_MSC_VER)
#pragma warning(pop)
#endif

void OpenSBCDaemon::OnConfigChanged( PThread & /*thrd*/, INT data )
{
  OSSAppConfig * _config = reinterpret_cast<OSSAppConfig*>(data);

  if( _config != NULL )
  {
    OSSAppConfig & config = *_config;
    PTRACE( 1, "Configuration change detected" ); 

    int logLevel = config.GetInteger( configKeySection, configKeyAppLogLevel, 3 );
    OString syslogAddress = config.GetString( configKeySection, 
      configKeySysLogServer, "127.0.0.1" );
    
    OString logPrefix = config.GetString( 
      configKeySection, configKeyAppLogFilePrefix, "b2bua" );

    if( !m_InstanceId.IsEmpty() )
      logPrefix  += "-" + m_InstanceId;

    Logger::SetDefaultLevel( logLevel );
    PTRACE( 1, "Setting SIP Log Level to " << logLevel );
    

    int ptraceLevel = config.GetInteger( configKeySection, configKeyCoreLogLevel, 1 );
    SetPTraceLevel( ptraceLevel );
    PTRACE( 1, "Setting PTRACE Log Level to " << ptraceLevel );

    BOOL mediaProxy = config.GetBoolean( configKeySection, configKeyMediaProxy, FALSE );
    if( m_EndPoint )
      m_EndPoint->SetMediaProxyIfPrivate( !mediaProxy );

    if( m_SBC != NULL )
      m_SBC->OnConfigChanged( config );

    m_IsInitialized = TRUE;

    static BOOL firstRun = TRUE;
    
    
    if( firstRun && m_SBC->GetUAMode() == B2BUserAgent::ProxyOnlyMode && 
          config.GetType() == OSSAppConfig::XMLRPCClientProvider )
    {

      OString serverURL = GetExtendedArgs().GetParameter( "xml-rpc-server" );
      if( !serverURL.IsEmpty() )
      {
        PArray<PStringToString>  cmd;
        PArray<PStringToString>  response;
        PStringToString * cmdParams = new PStringToString();
        cmdParams->SetAt( "iid", GetExtendedArgs().GetParameter( "iid" ) );
        cmd.SetAt( 0, cmdParams );
        config.SendRequest( "Proxy.Startup", cmd, response, serverURL.c_str() );
      }
    }else if( firstRun && m_SBC->GetUAMode() == B2BUserAgent::B2BOnlyMode && 
          config.GetType() == OSSAppConfig::XMLRPCClientProvider )
    {

      OString serverURL = GetExtendedArgs().GetParameter( "xml-rpc-server" );
      if( !serverURL.IsEmpty() )
      {
        PArray<PStringToString>  cmd;
        PArray<PStringToString>  response;
        PStringToString * cmdParams = new PStringToString();
        cmdParams->SetAt( "iid", GetExtendedArgs().GetParameter( "iid" ) );
        cmd.SetAt( 0, cmdParams );
        config.SendRequest( "B2BUA.Startup", cmd, response, serverURL.c_str() );
      }
    }
    firstRun = FALSE;
  } 

 
}


void OpenSBCDaemon::OnHandleCommandArgs(
    PINDEX i,
    const PString & arg
)
{
  if( i == 0 )  /// 'X'
  {
    m_HasExtendedArgs = TRUE;
    m_ExtendedArgs = (const char *)arg;
    PTRACE( 1, m_ExtendedArgs.AsString() );
  }
}

void OpenSBCDaemon::SetPTraceLevel( unsigned level )
{
#if PTRACING
  PTrace::SetLevel( level );
  PTrace::ClearOptions( PTrace::DateAndTime );
  PTrace::SetOptions( PTrace::Timestamp );
#endif
}

//////////////////////////////////////////////////////////////////////////////



OpenSBC::OpenSBC( 
  OpenSBCDaemon * application,
  UAMode mode
  ) : B2BUserAgent( mode )
{
  m_Application = application;
  
  m_SBCAuthHandler = NULL;
  m_SBCCallHandler = NULL;
  m_SBCRoutingHandler = NULL;
  m_SBCIVRHandler = NULL;
  m_AcceptAllCalls = FALSE;
  m_AcceptAllRegistration = FALSE;
}

void OpenSBC::OnStart( 
  OSSAppConfig & 
)
{
}

void OpenSBC::OnStop()
{
}

void OpenSBC::OnConfigChanged( 
  OSSAppConfig & config 
)
{
  int workerThreadCount = config.GetInteger( configKeySection, configKeyTransactionThreadCount, 10 );

  SIPURI bindAddress;
  OStringArray moreBindAddress;
  if( !m_Application->HasExtendedArgs() )
  {
    PINDEX listSize = config.GetListSize( configKeySection, configKeyInterfaceAddress );
    if( listSize == 0 )
      bindAddress  = "sip:*:5060";
    else
    {
      bindAddress = (const char *)config.GetListItem( configKeySection, configKeyInterfaceAddress, 0 );
      if( bindAddress.GetHost().IsEmpty() )
        bindAddress = "sip:*:5060" ;

      /// check if there are more interfaces configured
      if( listSize > 1 )
      {
        for( PINDEX i = 1; i < listSize; i++ )
        {
          OString moreIface = config.GetListItem( configKeySection, configKeyInterfaceAddress, i );
          if( !moreIface.IsEmpty() )
            moreBindAddress.AppendString( moreIface );
        }
      }
    }
  }else
  {
    bindAddress = m_Application->GetExtendedArgs().AsString( FALSE );
  }

  B2BUAConnection::SetSeizeTimeout( config.GetInteger( configKeySection, configKeySeizeTimeout, 0 ) );
  B2BUAConnection::SetAlertingTimeout( config.GetInteger( configKeySection, configKeyAlertingTimeout, 0 ) );
  
  // Indicate if we will accept all registrations
  m_AcceptAllRegistration = config.GetBoolean( configKeyAccountSection, configKeyAcceptAllRegistration, 0 );
  GetRegistrar()->SetAcceptAllRegistrations( m_AcceptAllRegistration );

  // Enable Feature SIP Session Timers
  GetB2BUAEndPoint()->EnableSessionTimer( config.GetBoolean( configKeySection, configKeySIPSessionTimer, 0 ) );

  Encryption::Engine::m_Key = (const char *)config.GetString( configKeySection, configKeyEncryptionKey, Encryption::Engine::m_DefKey );
  OString encMode = config.GetString( configKeySection, configKeyEncryption, "XOR" );
  if( encMode *= "XOR" )
    Encryption::Engine::m_Type = Encryption::Engine::Type_XOR;
  else if( encMode *= "DiffShift" )
    Encryption::Engine::m_Type = Encryption::Engine::Type_DiffShift;

  /// create the domain account database
  PStringArray accounts;
  for( PINDEX i = 0; i <  config.GetListSize( configKeyAccountSection, configKeyAccount ); i++ )
    accounts.AppendString( config.GetListItem( configKeyAccountSection, configKeyAccount, i ) );
  AddLocalDomainAccounts( accounts, TRUE );

  m_AcceptAllCalls = config.GetBoolean( configKeyTrustedDomainSection, configKeyAcceptAllCalls, 0 );

  /// create the domain account database
  PStringArray domains;
  for( PINDEX i = 0; i <  config.GetListSize( configKeyTrustedDomainSection, configKeyDomain ); i++ )
    domains.AppendString( config.GetListItem( configKeyTrustedDomainSection, configKeyDomain, i ) );
  AddTrustedDomains( domains, TRUE );

  /// chech whether we should enable local Refer
  BOOL localRefer = config.GetBoolean( configKeySection, 
    configKeyEnableLocalRefer, 0 );
  GetB2BUAEndPoint()->EnableLocalRefer( localRefer );

  PStringArray privacyDomains;
  for( PINDEX i = 0; i <  config.GetListSize( configKeyPrivacyTrustedDomainSection, configKeyPrivacyDomain ); i++ )
    privacyDomains.AppendString( config.GetListItem( configKeyPrivacyTrustedDomainSection, configKeyPrivacyDomain, i ) );
  AddPrivacyTrustedDomains( privacyDomains, TRUE );
  SetPrivacyDefaultRealm( config.GetString( configKeyPrivacyTrustedDomainSection, configKeyPrivacyDefaultRealm, "" ) );

  PIPSocket::Address iface = PIPSocket::Address( bindAddress.GetHost() );
  if( !iface.IsValid() )
    iface = 0;

  WORD port = 5060;
  if( bindAddress.GetPort() != "0" )
    port = (WORD)bindAddress.GetPort().AsUnsigned();

  GetDefaultProfile().GetTransportProfile().EnableUDP( iface, port  );

  if( moreBindAddress.GetSize() > 0 )
    GetDefaultProfile().GetTransportProfile().SetUDPAdditionalListenerAddress( moreBindAddress );

  OString staticRTPMediaAddress = config.GetString( configKeySection, 
    configKeyStaticRTPMediaAddress, "" );

#if ENABLE_TCP_TRANSPORT
  GetDefaultProfile().GetTransportProfile().EnableTCP( iface, port  );
#endif

  Initialize( workerThreadCount );

  ((SBCRoutingHandler*)m_RoutingHandler)->RefreshStaticRoutes();

  m_MediaHandler->SetStaticMediaAddress( staticRTPMediaAddress );


  OString defaultVXML = "<?xml version=\"1.0\"?>"
                "<vxml version=\"1.0\">"
                  "<form id=\"root\">"
                    "<audio src=\"welcome.wav\">"
                      "Welcome to the auto attendant main menu."
                    "</audio>"
                  "</form>"
                "</vxml>";

  OString vxmlScript = config.GetString( configAutoAttendantSection, configKeyVXMLScript, defaultVXML );
  GetMediaServer()->GetMediaServerEndPoint()->SetDefaultVXML( vxmlScript );
}

BOOL OpenSBC::IsTrustedDomain( const SIPURI & uri )const
{
  PWaitAndSignal lock( m_TrustedDomainsMutex );
  OString key( uri.GetHost() );
  PINDEX index = m_TrustedDomains.GetStringsIndex(key);
  return index != P_MAX_INDEX;
}

BOOL OpenSBC::AddTrustedDomains(
  const PStringArray & domains,
  BOOL dumpExisting
)
{
  PWaitAndSignal lock( m_TrustedDomainsMutex );
  
  if( dumpExisting )
    m_TrustedDomains.RemoveAll();

  for( PINDEX i = 0; i < domains.GetSize(); i++ )
  {
    SIPURI domain( (const char *)domains[i] );
    OString key = domain.GetHost();
    m_TrustedDomains.AppendString( key );
  }

  return TRUE;
}


BOOL OpenSBC::IsPrivacyTrustedDomain( 
  const SIPURI & uri,
  SIPURI & domainEntry
)const
{
  PWaitAndSignal lock( m_PrivacyTrustedDomainsMutex );
  OString key( uri.GetHost() );
  PINDEX index = m_PrivacyTrustedDomains.GetStringsIndex(key);
  if( index != P_MAX_INDEX )
  {
    OString * str = dynamic_cast<OString*>(m_PrivacyTrustedDomains.GetAt(index));
    if( str != NULL )
      domainEntry = *str;

    return TRUE;
  }

  return FALSE;
}

BOOL OpenSBC::AddPrivacyTrustedDomains(
  const PStringArray & domains,
  BOOL dumpExisting
)
{
  PWaitAndSignal lock( m_PrivacyTrustedDomainsMutex );
  
  if( dumpExisting )
    m_PrivacyTrustedDomains.RemoveAll();

  for( PINDEX i = 0; i < domains.GetSize(); i++ )
  {
    SIPURI domain( (const char *)domains[i] );
    OString key = domain.GetHost();
    m_PrivacyTrustedDomains.AppendString( key );
  }

  return TRUE;
}

BOOL OpenSBC::FindLocalDomainAccount(
  const SIPURI & uri,
  SIPURI & accountURI
)const
{
  PWaitAndSignal lock( m_LocalDomainAccountsMutex );

  PStringStream key;
    key << uri.GetUser() << ":" << uri.GetHost();

  SIPURI * account = m_LocalDomainAccounts.GetAt( key );

  if( account == NULL )
    return FALSE;

  accountURI = *account;

  return TRUE;
}

BOOL OpenSBC::AddLocalDomainAccounts(
  const PStringArray & accounts,
  BOOL dumpExisting
)
{
  PWaitAndSignal lock( m_LocalDomainAccountsMutex );
  
  if( dumpExisting )
    m_LocalDomainAccounts.RemoveAll();

  for( PINDEX i = 0; i < accounts.GetSize(); i++ )
  {
    SIPURI accountURI( (const char *)accounts[i] );
    PStringStream key;
    key << accountURI.GetUser() << ":" << accountURI.GetHost();
    m_LocalDomainAccounts.SetAt( key, new SIPURI( accountURI ) );
  }

  return TRUE;
}

B2BCallInterface * OpenSBC::OnCreateCallInterface()
{
  if( m_SBCCallHandler == NULL )
    m_SBCCallHandler = new SBCCallHandler( *this );
  return m_SBCCallHandler;
}

B2BAuthInterface * OpenSBC::OnCreateAuthInterface()
{
  if( m_SBCAuthHandler == NULL  )
    m_SBCAuthHandler = new SBCAuthHandler( *this );
  return m_SBCAuthHandler;
}

B2BRoutingInterface * OpenSBC::OnCreateRoutingInterface()
{
  if( m_SBCRoutingHandler == NULL )
    m_SBCRoutingHandler = new SBCRoutingHandler( *this );
  return m_SBCRoutingHandler;
}

B2BMediaInterface * OpenSBC::OnCreateMediaInterface()
{
  return new B2BMediaInterface( *this );
}

B2BIVRInterface * OpenSBC::OnCreateIVRInterface()
{
  if( m_SBCIVRHandler == NULL )
    m_SBCIVRHandler = new SBCIVRHandler( *this );

  return m_SBCIVRHandler;
}



RegisterSession::AcceptRegistrationResponse OpenSBC::OnAcceptRegistration( 
  RegisterSession &,
  const SIPMessage &
)
{
  if( m_AcceptAllRegistration )
    return RegisterSession::AcceptByOk;

  return RegisterSession::AcceptByChallenge;
}

BOOL OpenSBC::OnIncomingCall(
  B2BUAConnection & connection,
  B2BUACall & call,
  SIPMessage & invite
)
{
  if( B2BUserAgent::OnIncomingCall( connection, call, invite ) )
  {
    if( connection.GetRoutes().IsEmpty() )
      return FALSE;

    SIPURI uri = connection.GetRoutes()[0];
    
    OString action;
    if( uri.GetParameter( "action", action ) )
    {
      if( action *= "redirect" )
      {
        SIPMessage msg;
        invite.CreateResponse( msg, SIPMessage::Code302_MovedTemporarily );
        msg.AppendContact( uri.GetBasicURI().AsString() );
        SendRequest( msg );

        connection.DestroyConnection( invite );
        return FALSE;
      }
    }

    return TRUE;
  }

  return FALSE;
}

//////////////////////////////////////////////////////////////////////////////



//////////////////////////////////////////////////////////////////////
/// Internal SIP Header

OpenSBC::UpstreamProxyRoute::UpstreamProxyRoute() 
  : SIPHeaderF("upstream-proxy-route")
{
}

OpenSBC::UpstreamProxyRoute::UpstreamProxyRoute(
  const SIPHeaderF & header
) : SIPHeaderF("upstream-proxy-route", header)
{
}

OpenSBC::UpstreamProxyRoute::UpstreamProxyRoute(
  const OString & header
) : SIPHeaderF("upstream-proxy-route", header)
{
}

OpenSBC::UpstreamProxyRoute::UpstreamProxyRoute(
  const MimeHeader & header
) : SIPHeaderF("upstream-proxy-route", header)
{
}

OpenSBC::UpstreamProxyRoute::UpstreamProxyRoute(
  const OStringArray & header
) : SIPHeaderF("upstream-proxy-route", header)
{
}

OpenSBC::UpstreamProxyRoute::UpstreamProxyRoute(
  const RouteURI::Collection & uri
) : SIPHeaderF("upstream-proxy-route", uri)
{
}

OpenSBC::UpstreamProxyRouteList::UpstreamProxyRouteList()
{
}

OpenSBC::UpstreamProxyRouteList::UpstreamProxyRouteList(
  const SIPHeaderFList & list
) : SIPHeaderFList( list )
{
}

OpenSBC::UpstreamProxyRouteList::UpstreamProxyRouteList(
  const SIPHeaderF::Collection & list
) : SIPHeaderFList( list )
{
}


////////////////////////////////////////////////////////////////////////////////

void OpenSBCDaemon::OnAdditionalHTTPResource(
  PHTTPSimpleAuth & authority
)
{
  httpNameSpace.AddResource(new SBCRegistrationStatusPage(*this, authority), PHTTPSpace::Overwrite);
  httpNameSpace.AddResource(new SBCResourceCounterPage(*this, authority), PHTTPSpace::Overwrite);
}

void OpenSBCDaemon::OnAdditionalHTTPLinks( 
  PHTML & html 
)
{
  html << PHTML::Paragraph() << "<center>" << PHTML::HotLink("OpenSBC Registration Status") << "OpenSBC Registration Status" << PHTML::HotLink();
  html << PHTML::Paragraph() << "<center>" << PHTML::HotLink("OpenSBC Resource Counters") << "OpenSBC Resource Counters" << PHTML::HotLink();
}

static void SpliceMacro(PString & text, const PString & token, const PString & value)
{

  PRegularExpression RegEx("<?!--#status[ \t\r\n]+" + token + "[ \t\r\n]*-->?",
                           PRegularExpression::Extended|PRegularExpression::IgnoreCase);
  PINDEX pos, len;
  while (text.FindRegEx(RegEx, pos, len))
    text.Splice(value, pos, len);
}


PCREATE_SERVICE_MACRO_BLOCK(SBCRegInfo,P_EMPTY,P_EMPTY,htmlBlock)
{
  OpenSBCDaemon * app = dynamic_cast<OpenSBCDaemon *>( OSSApplication::GetInstance() );
  return app->OnLoadRegistrationStatus( htmlBlock );
}

PString OpenSBCDaemon::OnLoadRegistrationStatus( 
  const PString & htmlBlock 
)
{
  RegistrationDatabase & regDB = GetSBC()->GetRegistrar()->GetRegistrationDB();
  PWaitAndSignal lock( regDB.GetMutex() );
  OString substitution;

  for( PINDEX i = 0; i < regDB.GetSize(); i++ )
  {
    CachePointer * cacheData = regDB.CreatePointer( i );
    if( cacheData != NULL )
    {
      SIPMessage * reg = dynamic_cast<SIPMessage *>(cacheData->GetObject());
      if( reg != NULL )
      {
        PString insert = htmlBlock;
        //PString inst = insert.c_str();
        SpliceMacro(insert, "URI", (*cacheData)->GetObjectIdentifier() );
        SpliceMacro(insert, "AOR", reg->GetContactTopURI().AsString( FALSE ) );
        SpliceMacro(insert, "Expires", OString( (*cacheData)->GetLifeSpan() ) );
        SpliceMacro(insert, "Call-ID", reg->GetCallId());
        substitution+=(const char *)insert;
      }

      delete cacheData;
    }
  }

  return substitution;
}

BOOL SBCRegistrationStatusPage::Post(PHTTPRequest & request,
                          const PStringToString & data,
                          PHTML & msg)
{
  
  PTRACE(2, "Main\tClear call POST received " << data);

  msg << PHTML::Title() << "Accepted Control Command" << PHTML::Body()
      << PHTML::Heading(1) << "Accepted Control Command" << PHTML::Heading(1);

  app.OnPostControl(data, msg);


  msg << PHTML::Paragraph()
      << PHTML::HotLink(request.url.AsString()) << "Reload page" << PHTML::HotLink()
      << "&nbsp;&nbsp;&nbsp;&nbsp;"
      << PHTML::HotLink("/") << "Home page" << PHTML::HotLink();

  PServiceHTML::ProcessMacros(request, msg, "html/status.html",
                              PServiceHTML::LoadFromFile|PServiceHTML::NoSignatureForFile);
                              
  return TRUE;
}

BOOL OpenSBCDaemon::OnPostControl(const PStringToString & data, PHTML & msg)
{
  BOOL ok = FALSE;
  for( PINDEX i = 0; i < data.GetSize(); i++ )
  {
    OString key = data.GetKeyAt(i);
    OString command = data.GetDataAt(i);

    if (command == "Unregister") 
    {
      SIPSession::GCRef session = GetSBC()->GetRegistrar()->FindGCRefByCallId( key );
      if( session != NULL )
      {
        ok = TRUE;
        session->Destroy();

        msg << PHTML::Heading(2) << "AOR removed for Call-ID " << key << PHTML::Heading(2);

      }else
      {
        msg << PHTML::Heading(2) << "No Registration Found for Call-ID " << key  << PHTML::Heading(2);
      }
    }
  }
  return ok;
}

void OpenSBCDaemon::OnAdditionalXMLRPCMethods(
  PStringArray & methods
)
{
  methods.AppendString( "Registrar.GetRegStatus" );
  methods.AppendString( "Registrar.FindRegistration" );
  methods.AppendString( "Registrar.Unregister" );
  methods.AppendString( "Routing.AddRelayRoute" );
  methods.AppendString( "Proxy.Startup" );
  methods.AppendString( "B2BUA.Startup" );
}

void OpenSBCDaemon::OnCustomXMLRPCRequest(
  const PString & method,
  PArray<PStringToString> & request,
  PArray<PStringToString> & response
)
{
  if( !m_IsInitialized )
  {
    PStringToString * immediateResponse = new PStringToString();
    immediateResponse->SetAt( "ResponseCode", "999" );
    immediateResponse->SetAt( "ResponseMessage", "Server Not Ready" );
    response.SetAt( 0, immediateResponse );
    return;
  }

  PStringToString * requestParams = dynamic_cast<PStringToString*>(request.GetAt(0));

  if( method == "Registrar.GetRegStatus" )
  {
    RegistrationDatabase & regDB = GetSBC()->GetRegistrar()->GetRegistrationDB();
    PWaitAndSignal lock( regDB.GetMutex() );
    OString substitution;

    PStringToString * immediateResponse = new PStringToString();
    immediateResponse->SetAt( "ResponseCode", "0" );
    immediateResponse->SetAt( "ResponseMessage", "Ok" );
    response.SetAt( 0, immediateResponse );
    
    for( PINDEX i = 0; i < regDB.GetSize(); i++ )
    {
      CachePointer * cacheData = regDB.CreatePointer( i );
      if( cacheData != NULL )
      {
        SIPMessage * reg = dynamic_cast<SIPMessage *>(cacheData->GetObject());
        if( reg != NULL )
        {
          PStringToString * record = new PStringToString();
          record->SetAt( "URI", (*cacheData)->GetObjectIdentifier() );
          record->SetAt( "AOR", reg->GetContactTopURI().AsString( FALSE ) );
          record->SetAt( "Expires", OString( (*cacheData)->GetLifeSpan() ) );
          record->SetAt( "Call-ID", reg->GetCallId());
          response.SetAt( i + 1, record );
        }
        delete cacheData;
      }
    }
  }else if( method == "Registrar.FindRegistration" )
  {
    PTRACE( 1, "Registrar.FindRegistration" );
    RegistrationDatabase & regDB = GetSBC()->GetRegistrar()->GetRegistrationDB();
    PWaitAndSignal lock( regDB.GetMutex() );

    if( requestParams == NULL )
    {
      PStringToString * immediateResponse = new PStringToString();
      immediateResponse->SetAt( "ResponseCode", "999" );
      immediateResponse->SetAt( "ResponseMessage", "Missing URI parameter" );
      response.SetAt( 0, immediateResponse );
      return;
    }else
    {
      PString * uri = requestParams->GetAt( "URI" );
      if( uri == NULL || uri->IsEmpty() )
      {
        PStringToString * immediateResponse = new PStringToString();
        immediateResponse->SetAt( "ResponseCode", "999" );
        immediateResponse->SetAt( "ResponseMessage", "Missing URI parameter" );
        response.SetAt( 0, immediateResponse );
        return;
      }

      SIPURI userURI( (const char *)*uri );
      SIPMessage registration;
      PStringToString * immediateResponse = new PStringToString();
      if( regDB.FindRegistration( userURI, registration ) )
      {
        SIPURI targetURI;
        GetSBC()->GetSBCRoutingHandler()->ExtractTargetURI( targetURI, registration );
        immediateResponse->SetAt( "Registration", targetURI.AsString() );
        immediateResponse->SetAt( "ResponseCode", "0" );
        immediateResponse->SetAt( "ResponseMessage", "Ok" );
        response.SetAt( 0, immediateResponse );
        PTRACE( 1, "Found registration " << targetURI );
        return;
      }else
      {
        immediateResponse->SetAt( "ResponseCode", "0" );
        immediateResponse->SetAt( "ResponseMessage", "No Registation" );
        response.SetAt( 0, immediateResponse );
        return;
      }
    }
  }else if( method == "Routing.AddRelayRoute" )
  {
    PStringToString * immediateResponse = new PStringToString();
    if( requestParams == NULL )
    { 
      immediateResponse->SetAt( "ResponseCode", "999" );
      immediateResponse->SetAt( "ResponseMessage", "Missing Route parameter" );
      response.SetAt( 0, immediateResponse );
      return;
    }else
    {
      PString * route = requestParams->GetAt( "Route" );
      if( route == NULL || route->IsEmpty() )
      {
        immediateResponse->SetAt( "ResponseCode", "999" );
        immediateResponse->SetAt( "ResponseMessage", "Missing Route parameter" );
        response.SetAt( 0, immediateResponse );
        return;
      }

      GetSBC()->GetSBCRoutingHandler()->AppendAppRelayRoutes( *route );
      immediateResponse->SetAt( "ResponseCode", "0" );
      immediateResponse->SetAt( "ResponseMessage", "Ok" );
      response.SetAt( 0, immediateResponse );
      return;
    }
  }else if( method == "Registrar.Unregister" )
  {
    PString * key = requestParams->GetAt( "Call-ID" );
    PStringToString * immediateResponse = new PStringToString();
    if( key == NULL || key->IsEmpty() )
    {
      immediateResponse->SetAt( "ResponseCode", "999" );
      immediateResponse->SetAt( "ResponseMessage", "Missing Call-ID parameter" );
      response.SetAt( 0, immediateResponse );
      return;
    }

    SIPSession::GCRef session = GetSBC()->GetRegistrar()->FindGCRefByCallId( (const char *)*key );
    OStringStream msg;
    if( session != NULL )
    {
      session->Destroy();
      msg << "AOR removed for Call-ID " << *key ;
    }else
    {
      msg << "No Registration Found for Call-ID " << *key ;
    }

    immediateResponse->SetAt( "ResponseCode", "0" );
    immediateResponse->SetAt( "ResponseMessage", "Ok" );
    response.SetAt( 0, immediateResponse );

    PStringToString * message = new PStringToString();
    message->SetAt( "Message", msg.str().c_str() );
    response.SetAt( 1, message );
    return;
  }
}

const PString& OpenSBCDaemon::GetName() const
{
  if( m_ExtendedAppName.IsEmpty() )
    return PProcess::GetName();
  return m_ExtendedAppName;
}

SBCRegistrationStatusPage::SBCRegistrationStatusPage(OpenSBCDaemon & _app, PHTTPAuthority & auth)
  : PServiceHTTPString("OpenSBC Registration Status", "", "text/html; charset=UTF-8", auth),
    app(_app)
{
  
  PHTML html;

  html << PHTML::Title("OpenSBC Registration Status")
       << "<meta http-equiv=\"Refresh\" content=\"30\">\n"
       << PHTML::Body()
       << app.GetPageGraphic()

       << PHTML::Paragraph()
       << PHTML::HotLink("OpenSBC Registration Status") << "Reload page" << PHTML::HotLink()
       << "&nbsp;&nbsp;&nbsp;&nbsp;"
       << PHTML::HotLink("/") << "Home page" << PHTML::HotLink()

       << PHTML::Paragraph() /*<< "<center>"*/
       << PHTML::Form("POST")
       << PHTML::TableStart("border=1")
       << PHTML::TableRow()
       << PHTML::TableHeader()
       << "&nbsp;URI&nbsp;"
       << PHTML::TableHeader()
       << "&nbsp;AOR&nbsp;"
       << PHTML::TableHeader()
       << "&nbsp;Expires&nbsp;"
       << PHTML::TableHeader()
       << "&nbsp;Call-ID&nbsp;"
       << "<!--#macrostart SBCRegInfo-->"
         << PHTML::TableRow()
         << PHTML::TableData()
         << "<!--#status URI-->"
         << PHTML::TableData()
         << "<!--#status AOR-->"
         << PHTML::TableData()
         << "<!--#status Expires-->"
         << PHTML::TableData()
         << "<!--#status Call-ID-->"
         << PHTML::TableData()
         << PHTML::SubmitButton("Unregister", "!--#status Call-ID--")
       << "<!--#macroend SBCRegInfo-->"
       << PHTML::TableEnd()

       
       << PHTML::Form()
       << PHTML::HRule()

       << app.GetCopyrightText()
       << PHTML::Body();

  string = html;

}

PCREATE_SERVICE_MACRO_BLOCK(SBCCounters,P_EMPTY,P_EMPTY,htmlBlock)
{
  OpenSBCDaemon * app = dynamic_cast<OpenSBCDaemon *>( OSSApplication::GetInstance() );
  return app->OnLoadResourceCounters( htmlBlock );
}

SBCResourceCounterPage::SBCResourceCounterPage(OpenSBCDaemon & _app, PHTTPAuthority & auth)
  : PServiceHTTPString("OpenSBC Resource Counters", "", "text/html; charset=UTF-8", auth),
                       app(_app)
{
  
  PHTML html;

  html << PHTML::Title("OpenSBC Resource Counters")
      << "<meta http-equiv=\"Refresh\" content=\"1\">\n"
      << PHTML::Body()
      << app.GetPageGraphic()

      << PHTML::Paragraph()
      << PHTML::HotLink("OpenSBC Resource Counters") << "Reload page" << PHTML::HotLink()
      << "&nbsp;&nbsp;&nbsp;&nbsp;"
      << PHTML::HotLink("/") << "Home page" << PHTML::HotLink()

      << PHTML::Paragraph() /*<< "<center>"*/
      << PHTML::Form("POST")
      << PHTML::TableStart("border=1")
      << PHTML::TableRow()
      
      << PHTML::TableHeader()
      << "&nbsp;ICT&nbsp;"
      << PHTML::TableHeader()
      << "&nbsp;NICT&nbsp;"
      << PHTML::TableHeader()
      << "&nbsp;IST&nbsp;"
      << PHTML::TableHeader()
      << "&nbsp;NIST&nbsp;"
      << PHTML::TableHeader()
      << "&nbsp;CallSession&nbsp;"
      << PHTML::TableHeader()
      << "&nbsp;Connection&nbsp;"
      << PHTML::TableHeader()
      << "&nbsp;Registration&nbsp;"
      << PHTML::TableHeader()
      << "&nbsp;RTPSession&nbsp;"
      << PHTML::TableHeader()
      << "&nbsp;EventQueue&nbsp;"
      << PHTML::TableHeader()
      << "&nbsp;Cache&nbsp;"
      << PHTML::TableHeader()
      << "&nbsp;Garbage&nbsp;"
  
      << "<!--#macrostart SBCCounters-->"
      << PHTML::TableRow()
      
      << PHTML::TableData()
      << "<!--#status ICT-->"

      << PHTML::TableData()
      << "<!--#status NICT-->"

      << PHTML::TableData()
      << "<!--#status IST-->"

      << PHTML::TableData()
      << "<!--#status NIST-->"

      << PHTML::TableData()
      << "<!--#status CallSession-->"

      << PHTML::TableData()
      << "<!--#status Connection-->"

      << PHTML::TableData()
      << "<!--#status Registration-->"

      << PHTML::TableData()
      << "<!--#status RTPSession-->"

      << PHTML::TableData()
      << "<!--#status EventQueue-->"

      << PHTML::TableData()
      << "<!--#status Cache-->"

      << PHTML::TableData()
      << "<!--#status Garbage-->"
      
      << "<!--#macroend SBCCounters-->"
      << PHTML::TableEnd()

       
      << PHTML::Form()
      << PHTML::HRule()

      << app.GetCopyrightText()
      << PHTML::Body();

  string = html;

}

PString OpenSBCDaemon::OnLoadResourceCounters( 
    const PString & htmlBlock 
                                               )
{

  PString substitution;
  PString insert = htmlBlock;
     
  
  SpliceMacro(insert, "ICT", (int)ResourceCounter::m_RC_ICT );
  SpliceMacro(insert, "NICT", (int)ResourceCounter::m_RC_NICT);
  SpliceMacro(insert, "IST", (int)ResourceCounter::m_RC_IST);
  SpliceMacro(insert, "NIST", (int)ResourceCounter::m_RC_NIST);
  SpliceMacro(insert, "CallSession", (int)ResourceCounter::m_RC_CallSession );
  SpliceMacro(insert, "Connection", (int)ResourceCounter::m_RC_Connection );
  SpliceMacro(insert, "Registration", (int)ResourceCounter::m_RC_Registration );
  SpliceMacro(insert, "RTPSession", (int)ResourceCounter::m_RC_RTPSession);
  SpliceMacro(insert, "EventQueue", (int)ResourceCounter::m_RC_EventQueue );
  SpliceMacro(insert, "Cache", (int)ResourceCounter::m_RC_Cache );
  SpliceMacro(insert, "Garbage", (int)ResourceCounter::m_RC_Garbage);
  
  
  substitution+=insert;
  

  return substitution;
}




