/*
 *
 * SBCAuthHandler.h
 * 
  * Open SIP Stack ( OSS )
 *
 * Copyright (c) opensipstack.org.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is OpenSIPStack.
 *
 * The Initial Developer of the Original Code is opensipstack.org.
 *
 * The author of this code is Joegen E. Baclor
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * The OpenSIPStack Library includes some GPL/LGPL code libraries that MAY be 
 * enabled at compile time using the --enable-gpllibs configure switch.  If
 * enabled, the content of this file is published under the terms of GNU 
 * General Public License Version 2.  For a detailed list of the available
 * GPL code, see $(opensipstack)/gnu/README file.
 *
 * $Log: SBCAuthHandler.h,v $
 * Revision 1.4  2007/07/14 07:53:56  joegenbaclor
 * Migrated code to use new STL based string class
 *
 * Revision 1.3  2007/07/07 15:21:17  joegenbaclor
 * Added iproute2 support for linux
 *
 * Revision 1.2  2007/02/16 11:09:20  joegenbaclor
 * More work on privacy
 *
 * Revision 1.1  2006/08/14 10:04:59  rcolobong
 * Convert B2BUA to SBC
 * Support MP logging
 *
 * Revision 1.2  2006/07/26 09:19:41  joegenbaclor
 * Added System Log support
 *
 * Revision 1.1  2006/07/11 14:34:38  joegenbaclor
 * Initial upload
 *
 *
 */

#ifndef SBCAUTHHANDLER_H
#define SBCAUTHHANDLER_H

#include "OpenSBC.h"
#include "B2BAuthInterface.h"


class SBCAuthHandler : public B2BAuthInterface
{
  PCLASSINFO( SBCAuthHandler, B2BAuthInterface );
public:

  SBCAuthHandler(  
    OpenSBC & b2bua 
  );

  virtual ~SBCAuthHandler();

  virtual B2BAuthResponse B2BAuthenticate(
    B2BUAConnection & connection,
    SIPMessage & invite,
    ProxyAuthenticate & auth
  );

  virtual void OnPostB2BAuthentication(
    B2BUAConnection & connection,
    SIPMessage & request,
    const SIPURI & uri
  );

  virtual BOOL RequireAuthorizationHash(
    const SIPURI & userURL,
    OString & a1,
    RegisterSession & session 
  );
};


#endif

