/*
 *
 * SBCIVRHandler.cxx
 * 
  * Open SIP Stack ( OSS )
 *
 * Copyright (c) opensipstack.org.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is OpenSIPStack.
 *
 * The Initial Developer of the Original Code is opensipstack.org.
 *
 * The author of this code is Joegen E. Baclor
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * The OpenSIPStack Library includes some GPL/LGPL code libraries that MAY be 
 * enabled at compile time using the --enable-gpllibs configure switch.  If
 * enabled, the content of this file is published under the terms of GNU 
 * General Public License Version 2.  For a detailed list of the available
 * GPL code, see $(opensipstack)/gnu/README file.
 *
 * $Log: SBCIVRHandler.cxx,v $
 * Revision 1.5  2007/07/07 15:21:17  joegenbaclor
 * Added iproute2 support for linux
 *
 * Revision 1.4  2007/06/15 09:12:07  joegenbaclor
 * Added IVR Transfer reject event
 *
 * Revision 1.3  2007/06/14 17:26:52  joegenbaclor
 * Changed grammar so that prompts may be interrupted anytime bu dtmf input
 *
 * Revision 1.2  2007/06/14 02:24:12  joegenbaclor
 * Forgot to check the password after FindLocalDomainAccount
 *
 * Revision 1.1  2007/06/13 09:22:01  joegenbaclor
 * Added IVR handler
 *
 *
 */

#include "SBCIVRHandler.h"
#include "SBCRoutingHandler.h"
#include "B2BUAConnection.h"

#define new PNEW

SBCIVRHandler::SBCIVRHandler( 
  OpenSBC & ua 
  ) : B2BIVRInterface( ua )
{
}

SBCIVRHandler::~SBCIVRHandler()
{
}

void SBCIVRHandler::OnIVREvent(
  B2BUAConnection * conn,
  const B2BIVREvent * evt
)
{
  OpenSBC & sbc = dynamic_cast<OpenSBC&>(GetB2BUA());

  if( evt == 0 )
    return;

  B2BIVREvent::Type type = evt->GetType();

  switch( type )
  {
    case B2BIVREvent::Event_IVROnOpenChannel:
      {
        conn->IVRPlayFile( "greeting", "prompts/basic/welcome.wav" );
        conn->IVRPlayFile( "enter-pin", "prompts/basic/enter_pin.wav" );
        conn->IVRSetDTMFGrammar( "pin-collect", 4, '#' );
      }
      break;
    case B2BIVREvent::Event_IVROnEndPlaying:
      {
        const OnEndPlaying * eventObject = static_cast<const OnEndPlaying *>(evt);
        PTRACE( 1, "Finished playing " << eventObject->m_Identifier );
        /*if( eventObject->m_Identifier == "enter-pin" )
          conn->IVRSetDTMFGrammar( "pin-collect", 4, '#' );
        else if( eventObject->m_Identifier == "enter-number" )
          conn->IVRSetDTMFGrammar( "number-collect", 20, '#' );
        else */if( eventObject->m_Identifier == "call-transfer" )
        {
          SIPHeader h("number-collect");
          if( conn->PopInternalHeader( "number-collect", h ) )
            conn->IVRTransferCall( h.GetHeaderBody(), FALSE );
        }
      }

      break;
    case B2BIVREvent::Event_IVROnUserInput:
      break;
    case B2BIVREvent::Event_IVROnEndCollection:
      {
        const OnEndCollection * eventObject = static_cast<const OnEndCollection *>(evt);
        PTRACE( 1, "Collected: " << eventObject->m_GrammarId << " : " << eventObject->m_GrammarBuffer );
        
        if( eventObject->m_GrammarId == "pin-collect" )
        {
          SIPURI userURI = conn->GetLeg1Call()->GetRemoteURI();
          SIPURI account;
          if( sbc.FindLocalDomainAccount( userURI, account ) )
          {
            if( account.GetPassword() == eventObject->m_GrammarBuffer )
            {
              conn->IVRPlayFile( "enter-number", "prompts/basic/enter_number.wav" );
              conn->IVRSetDTMFGrammar( "number-collect", 20, '#' );
            }else
            {
              conn->IVRPlayFile( "invalid-pin", "prompts/basic/invalid_pin.wav" );
              conn->IVRPlayFile( "enter-pin", "prompts/basic/enter_pin.wav" );
              conn->IVRSetDTMFGrammar( "pin-collect", 4, '#' );
            }
          }else
          {
            conn->IVRPlayFile( "invalid-pin", "prompts/basic/invalid_pin.wav" );
            conn->IVRPlayFile( "enter-pin", "prompts/basic/enter_pin.wav" );
            conn->IVRSetDTMFGrammar( "pin-collect", 4, '#' );
          }
        }else if( eventObject->m_GrammarId == "number-collect" )
        {
          SIPURI userURI = conn->GetLeg2Call()->GetRemoteURI();
          userURI.SetUser( eventObject->m_GrammarBuffer );
          if( sbc.GetSBCRoutingHandler()->GetStaticRoutes().HasRoute( userURI ) )
          {
            conn->IVRPlayFile( "call-transfer", "prompts/basic/call_transfer.wav" );
            conn->AddInternalHeader( "number-collect", eventObject->m_GrammarBuffer );
          }else
          {
            conn->IVRPlayFile( "invalid-number", "prompts/basic/invalid_number.wav" );
            conn->IVRPlayFile( "enter-number", "prompts/basic/enter_number.wav" );
            conn->IVRSetDTMFGrammar( "number-collect", 20, '#' );
            /// error message here
          }
        }
      }
      break;
    case B2BIVREvent::Event_IVROnTransferReject:
      break;
    default:
      break;
  }
}



