/*
 *
 * SBCRoutingHandler.cxx
 *
 * Open SIP Stack ( OSS )
 *
 * Copyright (c) opensipstack.org.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is OpenSIPStack.
 *
 * The Initial Developer of the Original Code is opensipstack.org.
 *
 * The author of this code is Joegen E. Baclor
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * The OpenSIPStack Library includes some GPL/LGPL code libraries that MAY be 
 * enabled at compile time using the --enable-gpllibs configure switch.  If
 * enabled, the content of this file is published under the terms of GNU 
 * General Public License Version 2.  For a detailed list of the available
 * GPL code, see $(opensipstack)/gnu/README file.
 *
 *
 * $Log: SBCRoutingHandler.cxx,v $
 * Revision 1.41  2007/07/15 12:27:26  joegenbaclor
 * Added capability to route to registered endpoints using static routes
 * Removed unused config parameters
 *
 * Revision 1.40  2007/07/14 07:53:56  joegenbaclor
 * Migrated code to use new STL based string class
 *
 * Revision 1.39  2007/07/07 15:21:17  joegenbaclor
 * Added iproute2 support for linux
 *
 * Revision 1.38  2007/07/06 14:18:25  rcolobong
 * Add Upstream Proxy Route feature
 *
 * Revision 1.37  2007/07/03 17:18:08  joegenbaclor
 * Got rid of compile warnings in unix builds
 *
 * Revision 1.36  2007/06/29 08:20:49  joegenbaclor
 * Removed constness for request param in B2BRouteRequest and RouteLocalRegistration to allow the target address to be set early for the request
 *
 * Revision 1.35  2007/06/18 15:30:54  joegenbaclor
 * Implemented Internal DNS Mapping
 *
 * Revision 1.34  2007/06/17 02:49:50  joegenbaclor
 * Introduced Global lock to PHTTPConfig
 *
 * Revision 1.33  2007/06/06 16:20:54  joegenbaclor
 * Commiting Media Server Related Changes
 *
 * Revision 1.32  2007/05/29 17:43:29  joegenbaclor
 * Initial work on auto attendant
 *
 * Revision 1.31  2007/05/18 12:17:09  joegenbaclor
 * Added DNS routing support for B2BOnlyMode
 *
 * Revision 1.30  2007/05/18 04:21:33  joegenbaclor
 * Set backdoor contact  for registers in B2BUpperRegMode
 *
 * Revision 1.29  2007/05/18 01:24:01  joegenbaclor
 * 1.  Added separate useragent to handle B2BUpperReg backdoor
 * 2.  Modified routing for B2BUpperReg
 * 3.  Corrected bug in ParserTools::GetCompactHeader()
 *
 * Revision 1.28  2007/05/15 12:18:50  joegenbaclor
 * allowed direct DNS routing of upper REGISTER in B2BUpperRegMode
 *
 * Revision 1.27  2007/05/14 06:24:49  joegenbaclor
 * Added DNS Failover and ICMP validation
 *
 * Revision 1.26  2007/04/10 06:32:00  joegenbaclor
 * Added new Encryption Mode
 *
 * Revision 1.25  2007/04/04 05:57:39  joegenbaclor
 * Fixed bug where user part of the r-uri is not set when routing through the registration database and the UA is behind NAT.
 *
 * Revision 1.24  2007/04/03 04:38:25  joegenbaclor
 * Made sure CALL_TRANSFER_PREFIX is stripped when routing through local registration database
 *
 * Revision 1.23  2007/03/28 09:09:21  joegenbaclor
 * Fixed bug in upper registration
 *
 * Revision 1.22  2007/03/22 09:09:39  joegenbaclor
 * Some logging aesthetics
 *
 * Revision 1.21  2007/03/08 01:51:52  joegenbaclor
 * Disallowed rewrite of startline for URI resolving to a remote host
 *
 * Revision 1.20  2007/02/16 11:09:20  joegenbaclor
 * More work on privacy
 *
 * Revision 1.19  2007/02/08 07:21:40  joegenbaclor
 * Corrected bug in RouteProxyRequest where calls to GetContactTopURI may return a NULL
 *  reference
 *
 * Revision 1.18  2007/02/07 01:14:16  joegenbaclor
 * sync laucher related codes
 *
 * Revision 1.17  2007/02/01 07:18:24  joegenbaclor
 * Brought back record route for FullMode
 *
 * Revision 1.16  2007/01/22 13:12:11  joegenbaclor
 * Updated VC710 project files
 *
 * Revision 1.15  2007/01/22 10:00:58  joegenbaclor
 * Fixed ProxyRouteRequest
 *
 * Revision 1.14  2007/01/19 08:20:29  joegenbaclor
 * Added check for none routable proxy request s
 *
 * Revision 1.13  2007/01/18 06:06:56  joegenbaclor
 * Incremented Build Number for latest development release
 *
 * Revision 1.12  2007/01/17 00:09:02  joegenbaclor
 * Added SysLog
 *
 * Revision 1.11  2007/01/15 10:06:57  joegenbaclor
 * More on XML-RPC Registration Queries
 *
 * Revision 1.10  2007/01/15 09:24:47  joegenbaclor
 * Added capability to find Registraton via XML-RPC
 *
 * Revision 1.9  2007/01/12 10:50:05  joegenbaclor
 * minor proxy bug fixes
 * more launcher code
 *
 * Revision 1.8  2007/01/11 10:00:26  joegenbaclor
 * more launcher code
 *
 * Revision 1.7  2007/01/10 05:00:39  joegenbaclor
 * Added B2BUpperReg mode.
 *
 * Revision 1.6  2006/12/21 08:56:59  joegenbaclor
 * Added pid in b2bua log file name
 * Fixed bug in proxy relay where startline URI is not rewritten if a static route is
 *  not found
 *
 * Revision 1.5  2006/12/14 02:18:37  rcolobong
 * Fix bug on OnRecordRouteShift causing an assertion failure during for loop
 *
 * Revision 1.4  2006/11/23 11:24:27  joegenbaclor
 * Added capability to route none INVITE requests like instant messages
 *
 * Revision 1.3  2006/11/22 11:41:01  rcolobong
 * 1. Update handling on Upper Registration
 * 2. Override method on OnOrphanedMessage
 * 3. Create new handling for route-shift in Record Routes Header
 *
 * Revision 1.2  2006/10/11 06:32:58  rcolobong
 * Seperate method for routing using local registration.
 *
 * Revision 1.1  2006/08/14 10:05:01  rcolobong
 * Convert B2BUA to SBC
 * Support MP logging
 *
 * Revision 1.7  2006/08/04 05:48:58  rcolobong
 * 1.  Add Getter for StaticRoutes
 * 2.  Add virtual to method RefreshStaticRoutes
 * 3.  RefreshStaticRoutes is not invoke during construction
 *
 * Revision 1.6  2006/07/19 02:19:13  joegenbaclor
 * 1.  Bug fixes concerning B2BUA destruction during during 3xx, 4xx, 5xx, 6xx local response
 * 2.  Fixed bug where Route header is never removed from outbound Invite resulting that may result to a loop when
 *      Inbound Invite has a Route header present
 * 3.  Added RemoveAllRoutes() in SIPMessage
 *
 * Revision 1.5  2006/07/13 06:53:03  joegenbaclor
 * 1.  Introduced Sanity checks for incoming SIP Message
 * 2.  Corrected bug in SIPURI where enum scheme is not getting recognized.
 * 3.  Added ENUM support to routing
 * 4.  Introduced global routing flag to indicate when and not to rewrite outbound To URI's
 *
 * Revision 1.4  2006/07/11 13:59:34  joegenbaclor
 * Completed Proxy functionalities for B2BUA
 *
 * Revision 1.3  2006/07/10 06:29:36  joegenbaclor
 * Completed crude Registration support for B2BUA
 *
 * Revision 1.2  2006/07/03 15:29:44  joegenbaclor
 * Started merging OpenB2BUA into OpenSBC
 *
 * Revision 1.1  2006/06/20 09:58:11  joegenbaclor
 * *** empty log message ***
 *
 *
 */

#include "SBCRoutingHandler.h"
#include "SIPTransportTools.h"

#ifdef P_LINUX
#if ENABLE_GPL_LIBS
#include "IPRoute.h"
#endif
#endif

#define new PNEW

static const char configKeySection[] = "B2BUA Routes";
static const char configKeyUpperRegistrarSection[] = "Upper Registration Routes";
static const char configKeyRelayRouteSection[] = "Relay Routes";
static const char configKeyDNSMapSection[] = "Internal DNS Mapping";
static const char configKeyRoute[] = "Route";
static const char configKeyMap[] = "Map";
static const char configKeyRewriteToURI[] = "Rewrite TO URI";
static const char configKeyB2BICMPValidation[] = "ICMP Validation";
static const char configKeyRelayICMPValidation[] = "ICMP Validation";
static const char configAutoAttendantSection[] = "Media Server";
static const char configKeyAutoAttendantNumber[] = "Media Server Number";
static const char configKeyUpstreamProxySection[] = "Upstream Proxy";
static const char configKeyUpstreamBackdoorURI[] = "Backdoor URI";
static const char configKeyUpstreamSendToBackdoor[] = "Send all calls to backdoor port";

SBCRoutingHandler::SBCRoutingHandler(  
  OpenSBC & b2bua 
) : B2BRoutingInterface( b2bua )
{
  m_UpstreamProxyRoutes.EnableMaskTargetURI();
}

SBCRoutingHandler::~SBCRoutingHandler()
{
}

void SBCRoutingHandler::RefreshStaticRoutes()
{
  OSSAppConfig * config = OSSAppConfig::GetInstance();
  
  /// lock the config to make sure noone can change data via http admin until we are done
  PWaitAndSignal configLock( PHTTPConfig::m_HTTPConfigMutex );
  
  PIPSocket::Address ivrListener = GetB2BUA().GetMediaServer()->GetListenerAddress();
  WORD ivrPort = GetB2BUA().GetMediaServer()->GetListenerPort();
  SIPURI ivrRoute( ivrListener, ivrPort );
  m_StaticRoutes.SetIVRRoute( ivrRoute );


  OStringStream autoAttendantRoute;
  autoAttendantRoute << "[sip:" 
                     << config->GetString( configAutoAttendantSection, configKeyAutoAttendantNumber, "5000" ) 
                     << "@*] ivr";
  
  OStringArray routes;
  routes.AppendString( autoAttendantRoute );
  for( PINDEX i = 0; i <  config->GetListSize( configKeySection, configKeyRoute ); i++ )
    routes.AppendString( config->GetListItem( configKeySection, configKeyRoute, i ) );
  
  if( routes.GetSize() > 0 )
    m_StaticRoutes.Parse( routes );

  m_StaticRoutes.SetICMPValidate( config->GetBoolean( 
        configKeySection, 
        configKeyB2BICMPValidation, FALSE ) );

  OStringArray upperRegistrationRoutes;
  for( PINDEX j = 0; j <  config->GetListSize( configKeyUpperRegistrarSection, configKeyRoute ); j++ )
    upperRegistrationRoutes.AppendString( config->GetListItem( configKeyUpperRegistrarSection, configKeyRoute, j ) );

  if( upperRegistrationRoutes.GetSize() > 0 )
    m_UpperRegistrationRoutes.Parse( upperRegistrationRoutes );

  OStringArray relayRoutes;
  for( PINDEX x = 0; x <  config->GetListSize( configKeyRelayRouteSection, configKeyRoute ); x++ )
    relayRoutes.AppendString( config->GetListItem( configKeyRelayRouteSection, configKeyRoute, x ) );

  if( relayRoutes.GetSize() > 0 )
    m_RelayRoutes.Parse( relayRoutes );

  m_RelayRoutes.SetICMPValidate( config->GetBoolean( 
        configKeyRelayRouteSection, 
        configKeyRelayICMPValidation, FALSE ) );
  
  OStringArray upstreamRoutes;
  for( PINDEX x = 0; x <  config->GetListSize( configKeyUpstreamProxySection, configKeyRoute ); x++ )
    upstreamRoutes.AppendString( config->GetListItem( configKeyUpstreamProxySection, configKeyRoute, x ) );

  if( upstreamRoutes.GetSize() > 0 )
    m_UpstreamProxyRoutes.Parse( upstreamRoutes );

  m_UpstreamBackdoorURI = config->GetString( configKeyUpstreamProxySection, 
    configKeyUpstreamBackdoorURI, "" );

  OStringArray dnsMap;
  for( PINDEX y = 0; y <  config->GetListSize( configKeyDNSMapSection, configKeyMap ); y++ )
    dnsMap.AppendString( config->GetListItem( configKeyDNSMapSection, configKeyMap, y ) );
  
  if( dnsMap.GetSize() > 0 )
    m_InternalDNSMap.Parse( dnsMap );

}

BOOL SBCRoutingHandler::B2BRouteRequest(
  SIPMessage & request,
  SIPURI  & route
)
{
  /// lock the config to make sure no one can change data via http admin until we are done
  PWaitAndSignal configLock( PHTTPConfig::m_HTTPConfigMutex );

  To to;
  request.GetTo( to );
  SIPURI uri = to.GetURI();
  OString user = uri.GetUser();
  OString callId = request.GetCallId();

  LOG_CONTEXT( LogInfo(), callId, "Routing URI " << uri << "through local registration database" );

  SIPMessage registration;

  if( !GetB2BUA().GetRegistrar()->GetRegistrationDB().FindRegistration( uri, registration ) )
  {
    LOG_CONTEXT( LogInfo(), callId, "*** NO ROUTE ** URI " << uri << "through local registration database" );
    return FALSE;
  }

  return RouteLocalRegistration( route, request, registration );
}



#ifndef CALL_TRANSFER_PREFIX 
#define CALL_TRANSFER_PREFIX "xfer-oss-"
#endif

BOOL SBCRoutingHandler::B2BRouteCall(
  B2BUAConnection & connection,
  SIPMessage & invite,
  BOOL ignoreRegistrations
)
{
  /// lock the config to make sure noone can change data via http admin until we are done
  PWaitAndSignal configLock( PHTTPConfig::m_HTTPConfigMutex );

  To to;
  invite.GetTo( to );
  SIPURI uri = to.GetURI();
  OString user = uri.GetUser();
  OString callId = invite.GetCallId();

  OString  transferPrefix = CALL_TRANSFER_PREFIX;

  if( user.Left( transferPrefix.GetLength() ) == transferPrefix )
  {
    LOG_CONTEXT( LogInfo(), callId, "Stripping Call Transfer Prefix for " << uri );
    user = user.Mid( transferPrefix.GetLength() );
   
    ///now get the refer-to from the uri 
    int newLen = user.Find( "-" );
    OString xferTo;
    if( newLen != P_MAX_INDEX )
    {
      if( newLen + 1 < user.GetLength() )
      {
        xferTo = user.Left( newLen );
        xferTo = PBase64::Decode( xferTo );
        user = user.Mid( newLen + 1 );
      }
    }

    uri.SetUser( user );
    
    ReferTo referTo( xferTo );
    invite.SetRequestURI( referTo.GetURI() );
    to.SetURI( uri );
    invite.SetTo( uri );
    
  }
  
  OSSAppConfig * config = OSSAppConfig::GetInstance();


  // this will check if we would 
  // use upstream proxy route on this call
  B2BUpstreamProxyRouteCall( connection, invite );

  SIPMessage registration;

  if( GetB2BUA().GetUAMode() == B2BUserAgent::B2BOnlyMode && 
      OSSAppConfig::GetInstance()->GetType() == OSSAppConfig::XMLRPCClientProvider 
  )
  {
    LOG_CONTEXT( LogInfo(), callId, "Fetching route through XML RPC for URI " << uri );
    OpenSBCDaemon * app = dynamic_cast<OpenSBCDaemon*>(OSSApplication::GetInstance());
    OString serverURL = app->GetExtendedArgs().GetParameter( "xml-rpc-registrar" );
    LOG_CONTEXT( LogInfo(), callId, "Using remote XML-RPC registrar at " << serverURL );
    BOOL ok = FALSE;
    if( !serverURL.IsEmpty() )
    {
      PArray<PStringToString>  cmd;
      PArray<PStringToString>  response;
      PStringToString * cmdParams = new PStringToString();
      cmdParams->SetAt( "URI", uri.AsString() );
      cmd.SetAt( 0, cmdParams );
      config->SendRequest( "Registrar.FindRegistration", cmd, response, serverURL.c_str() );
      PStringToString * responseParams = static_cast<PStringToString*>(response.GetAt(0));
      if( responseParams != NULL )
      {
        PString * reg = responseParams->GetAt( "Registration" );
        if( reg == NULL || reg->IsEmpty() )
        {
          LOG_CONTEXT( LogInfo(), callId, "*** NO ROUTE *** through XML RPC for URI " << uri );
          ok = FALSE;
        }else
        {
          LOG_CONTEXT( LogInfo(), callId, "*** ROUTE FOUND *** through XML RPC for URI " << uri );
          connection.AddRoute( SIPURI((const char *)*reg) );
          return TRUE;
        }
      }
    }

    if( !ok )
    {
      BOOL rewriteToURI = config->GetBoolean( 
        configKeySection, 
        configKeyRewriteToURI, FALSE );

      connection.SetRewriteToURI( rewriteToURI );

      return m_StaticRoutes.FindRoute( 
        *(GetB2BUA().GetStack().GetTransportManager()),
        GetB2BUA().GetRegistrar()->GetRegistrationDB(), 
        invite, 
        connection );
    }
  }else if( ignoreRegistrations || !GetB2BUA().GetRegistrar()->GetRegistrationDB().FindRegistration( uri, registration )  )
  {
    LOG_CONTEXT( LogInfo(), callId, "*** NO REGISTRATION FOUND *** Fetching route through local database URI " << uri );
    BOOL rewriteToURI = config->GetBoolean( 
      configKeySection, 
      configKeyRewriteToURI, FALSE );

    connection.SetRewriteToURI( rewriteToURI );

    BOOL found = m_StaticRoutes.FindRoute( 
      *(GetB2BUA().GetStack().GetTransportManager()),
        GetB2BUA().GetRegistrar()->GetRegistrationDB(), 
        invite, connection );

    if( !found )
    {
      LOG_CONTEXT( LogInfo(), callId, "*** NO STATIC ROUTE *** URI " << uri );
      if( ignoreRegistrations )
      {
        SIPURI routeURI;
        invite.GetRequestURI( routeURI );
        PIPSocket::Address verifiedAddress;
        WORD verifiedPort = 0;
        
        if( !GetB2BUA().GetStack().GetTransportManager()->IsLocalAddressAndPort( routeURI ) )
        {
          if( SIPTransport::Resolve( routeURI, verifiedAddress, verifiedPort, SIPTransport::UDP, FALSE ) )
          {
            LOG_CONTEXT( LogInfo(), callId, "*** Using Request URI *** " << routeURI << " as route" );
            connection.AddRoute( routeURI );
            return TRUE;
          }
        }else
        {
          routeURI = invite.GetToURI();
          if( !GetB2BUA().GetStack().GetTransportManager()->IsLocalAddressAndPort( routeURI ) )
          {
            if( SIPTransport::Resolve( routeURI, verifiedAddress, verifiedPort, SIPTransport::UDP, FALSE ) )
            {
              LOG_CONTEXT( LogInfo(), callId, "*** Using To URI *** " << routeURI << " as route" );
              connection.AddRoute( routeURI );
              return TRUE;
            }
          }
        }
      }else
      {
        BOOL found = RouteLocalRegistration( connection, invite, registration );
        if( !found )
        {
          SIPURI routeURI;
          invite.GetRequestURI( routeURI );
          PIPSocket::Address verifiedAddress;
          WORD verifiedPort = 0;
          
          if( !GetB2BUA().GetStack().GetTransportManager()->IsLocalAddressAndPort( routeURI ) )
          {
            if( SIPTransport::Resolve( routeURI, verifiedAddress, verifiedPort, SIPTransport::UDP, FALSE ) )
            {
              LOG_CONTEXT( LogInfo(), callId, "*** Using Request URI *** " << routeURI << " as route" );
              connection.AddRoute( routeURI );
              return TRUE;
            }
          }else
          {
            routeURI = invite.GetToURI();
            if( !GetB2BUA().GetStack().GetTransportManager()->IsLocalAddressAndPort( routeURI ) )
            {
              if( SIPTransport::Resolve( routeURI, verifiedAddress, verifiedPort, SIPTransport::UDP, FALSE ) )
              {
                LOG_CONTEXT( LogInfo(), callId, "*** Using To URI *** " << routeURI << " as route" );
                connection.AddRoute( routeURI );
                return TRUE;
              }
            }
          }
        }
      }
    }else
    {
      LOG_CONTEXT( LogInfo(), callId, "*** STATIC ROUTE FOUND *** URI " << uri );
    }
    return found;
  }

  return RouteLocalRegistration( connection, invite, registration );;
  
}


void SBCRoutingHandler::B2BUpstreamProxyRouteCall(
  B2BUAConnection & connection,
  SIPMessage & invite
)
{
  To to;
  invite.GetTo( to );
  SIPURI uri = to.GetURI();
  OString user = uri.GetUser();
  OString callId = invite.GetCallId();

  OString  transferPrefix = CALL_TRANSFER_PREFIX;
  if( user.Left( transferPrefix.GetLength() ) == transferPrefix )
  {
    LOG_CONTEXT( LogInfo(), callId, "Stripping Call Transfer Prefix for " << uri );
    user = user.Mid( transferPrefix.GetLength() );
    uri.SetUser( user );
  }

  BOOL found = m_UpstreamProxyRoutes.FindRoute( 
    *(GetB2BUA().GetStack().GetTransportManager()),
        GetB2BUA().GetRegistrar()->GetRegistrationDB(), 
        invite, connection, TRUE );

  if( found )
  {
    B2BRouteList & routes = connection.GetRoutes();
    B2BRouteList & upstreamProxyRoutes = connection.GetUpstreamProxyRoutes();

    // copy routes to upstream proxy routes
    for( PINDEX index = 0; index < routes.GetSize(); ++index )
      upstreamProxyRoutes.Append( new SIPURI( routes[index] ) );

    connection.RemoveRoutes(); // clean up all routes

    if( upstreamProxyRoutes.IsEmpty() )
      return;

    // specify the Upstream Backdoor URI
    if( m_UpstreamBackdoorURI.IsEmpty() )
      return;

    connection.SetUpstreamBackdoorURI( m_UpstreamBackdoorURI );

    // enable upstream proxy route
    LOG_CONTEXT( LogInfo(), callId, "*** UPSTREAM ROUTE ENABLED *** URI " << uri );
    connection.EnableUpstreamProxyRoute();
  } else
  {
    LOG_CONTEXT( LogInfo(), callId, "*** NO UPSTREAM ROUTE *** URI " << uri );
    connection.DisableUpstreamProxyRoute();
  }
}

ProxySession::RoutePolicy SBCRoutingHandler::RouteProxyRequest(
  ProxySession & session,
  SIPMessage & request
)
{
  /// lock the config to make sure no one can change data via http admin until we are done
  PWaitAndSignal configLock( PHTTPConfig::m_HTTPConfigMutex );

  //// check to TO header if it resolves to this proxy
  To to;
  request.GetTo( to );
  SIPURI target = to.GetURI();
  SIPURI targetURI = to.GetURI();
  OString callId = request.GetCallId();

  RequestLine requestLine;
  request.GetRequestLine( requestLine );
  SIPURI oldURI = requestLine.GetRequestURI();

  if( request.IsRegister() )
  {
    /// check if routable via upper registration
    LOG_CONTEXT( LogInfo(), callId, "Routing REGISTER for URI " << target );
    if( m_UpperRegistrationRoutes.FindRoute( request, targetURI, FALSE ) || 
      m_B2BUA.GetUAMode() == B2BUserAgent::FullMode ||
      m_B2BUA.GetUAMode() == B2BUserAgent::B2BUpperRegMode)
    {
      PIPSocket::Address targetAddress, listenerAddress;
      WORD targetPort, listenerPort = 5060;
      
      if( !SIPTransport::Resolve( targetURI, targetAddress, targetPort ) )
      {
        LOG_CONTEXT( LogError(), callId, "Unable to resolve Upper Registrar URI for  " << request.GetStartLine() ); 
        return ProxySession::RouteReject; 
      }

      if( !GetB2BUA().GetStack().GetTransportManager()->GetListenerAddress(
        SIPTransport::UDP, targetAddress, listenerAddress, listenerPort ) )
      {
        LOG_CONTEXT( LogError(), callId, "Unable to identify a network interface for  " << request.GetStartLine() );
        return ProxySession::RouteReject; 
      }

      SIPURI tempUri = request.GetContactTopURI();
      tempUri.SetHost( listenerAddress.AsString() );

      if( GetSBC().GetUAMode() == B2BUserAgent::FullMode || GetSBC().GetUAMode() == B2BUserAgent::B2BUpperRegMode )
        tempUri.SetPort( OString( GetSBC().GetBackDoorPort() ) );
      else
        tempUri.SetPort( OString( listenerPort ) );

      LOG_CONTEXT( LogInfo(), callId, "*** UPPER REGISTRATION ENABLE *** for " << target );
      session.AddInternalHeader( "enable-upper-registration", tempUri.AsString() );

      if( request.GetContactSize() > 0 )
      {
        Contact contact, newContact;
        request.GetContactAt( contact, 0 );
        ContactURI contactURI;
        contact.GetURI( contactURI );
        contactURI.SetURI( tempUri );
        newContact.AddURI( contactURI );
        request.RemoveAllContact();
        request.AppendContact( newContact );
      }

      if( !GetB2BUA().GetStack().GetTransportManager()->IsLocalAddressAndPort( target ) )
      {
        oldURI.SetHost( targetURI.GetHost() );
        oldURI.SetPort( targetURI.GetPort() );
        requestLine.SetRequestURI( oldURI );
        request.SetStartLine( requestLine );
      }

      return ProxySession::RouteBySession;
    }
  }else if( /*request.IsInvite()*/ !request.IsCancel() && !request.IsAck() )
  {
    BOOL relay = FALSE;
    BOOL allowRoundRobin = request.IsInvite();
   

    if( !m_AppRelayRoutes.FindRoute( request, targetURI, allowRoundRobin, TRUE ) )
    {
      if( m_RelayRoutes.FindRoute( request, targetURI, allowRoundRobin, TRUE ) )
        relay = TRUE;
    }else
    {
      relay = TRUE;
    }

    if( relay )
    {
      session.SetCreateDialog();
      session.SetDialogPeerAddress( targetURI.AsString() );

      RouteURI recordRoute;
      SIPURI routeURI;

      PIPSocket::Address targetAddress, listenerAddress;
      WORD targetPort = 0, listenerPort = 0;

      if( !SIPTransport::Resolve( targetURI, targetAddress, targetPort ) )
      {
        LOG_CONTEXT( LogInfo(), callId, "(2) Unable to resolve target UA for  " << request.GetStartLine() );
        return ProxySession::RouteReject;
      }

      if( !GetB2BUA().GetStack().GetTransportManager()->GetListenerAddress(
        SIPTransport::UDP, targetAddress, listenerAddress, listenerPort ) )
      {
        LOG_CONTEXT( LogError(), callId, "Unable to identify a network interface for  " << request.GetStartLine() );
        return ProxySession::RouteReject; 
      }

      routeURI.SetHost( listenerAddress.AsString() );
      routeURI.SetPort( OString( listenerPort ) );

      Contact contact;
      if( request.GetContactAt( contact, 0 ) )
      {
        ContactURI curi;
        if( contact.GetURI( curi ) )
        {
          if( SIPTransport::IsPrivateNetwork( listenerAddress ) && !SIPTransport::IsPrivateNetwork( curi.GetURI().GetAddress() ) )
          {
            /// public to private routing... lets tag the record route
            PIPSocket::Address shiftAddress;
            WORD shiftPort;

            GetB2BUA().GetStack().GetTransportManager()->GetListenerAddress(
              SIPTransport::UDP, request.GetTopContactURI().GetAddress(), shiftAddress, shiftPort );
            OString routeShift = shiftAddress.AsString() + ":" + OString( shiftPort );

            routeURI.AddParameter( "route-shift", routeShift );
          }
        }
      }


      if( m_B2BUA.GetUAMode() == B2BUserAgent::FullMode )
      {
        recordRoute.SetURI( routeURI );
        recordRoute.SetLooseRouter( TRUE );
        request.AppendRecordRoute( recordRoute );
      }

      if( !GetB2BUA().GetStack().GetTransportManager()->IsLocalAddressAndPort( target ) )
      {
        oldURI.SetHost( targetURI.GetHost() );
        oldURI.SetPort( targetURI.GetPort() );
        requestLine.SetRequestURI( oldURI );
        request.SetStartLine( requestLine );
      }

      return ProxySession::RouteBySession;
    }else
    {
      /// check if the request URI resolves to us
      RequestLine rq;
      if( request.GetRequestLine( rq ) )
      {
        if( session.GetTransportManager()->IsLocalAddressAndPort( rq.GetRequestURI() ) )
        {
          /// lets try to resolve the TO URI
          if( !session.GetTransportManager()->IsLocalAddressAndPort( request.GetToURI() ) )
          {
            SIPURI uri( request.GetToURI().AsString( FALSE, FALSE ) );
            uri.SetUser( rq.GetRequestURI().GetUser() );
            rq.SetRequestURI( uri );
            request.SetStartLine( rq );
          }else
          {
            return ProxySession::RouteReject;
          }
        }
      }
    }
  }else if(  ( request.IsCancel() || request.IsAck() ) )
  {
    targetURI = session.GetDialogPeerAddress();
    targetURI.SetUser( oldURI.GetUser() );
    requestLine.SetRequestURI( targetURI );
    request.SetStartLine( requestLine );

    /// se if we could resolve the resulting request
    PIPSocket::Address newTarget;
    WORD newPort = 0;
    if( SIPTransport::Resolve( request, newTarget, newPort ) )
    {
      if( session.GetTransportManager()->IsLocalAddressAndPort( newTarget, newPort ) )
      {
        /// request resolves to us ....
        session.SetSessionRejectResponse( SIPMessage::Code481_TransactionDoesNotExist );
        return ProxySession::RouteReject;
      } 
    }else
    {
      /// we are unable to resolve this address
      session.SetSessionRejectResponse( SIPMessage::Code481_TransactionDoesNotExist );
      return ProxySession::RouteReject;
    }

  }else if( !request.IsRequest() )
  {
    OnRecordRouteShift( request );
  }else
  {
    return RouteProxyNISTRequest( session, request );
  }
    
  return ProxySession::RouteBySession;
}

ProxySession::RoutePolicy SBCRoutingHandler::RouteProxyNISTRequest(
  ProxySession & session,
  SIPMessage & request
)
{
  /// lock the config to make sure no one can change data via http admin until we are done
  PWaitAndSignal configLock( PHTTPConfig::m_HTTPConfigMutex );

  /// se if we could resolve the resulting request
  PIPSocket::Address newTarget;
  WORD newPort = 0;
  if( SIPTransport::Resolve( request, newTarget, newPort ) )
  {
    if( session.GetTransportManager()->IsLocalAddressAndPort( newTarget, newPort ) )
    {
      /// request resolves to us ....
      session.SetSessionRejectResponse( SIPMessage::Code481_TransactionDoesNotExist );
      return ProxySession::RouteReject;
    }
  }else
  {
    session.SetSessionRejectResponse( SIPMessage::Code404_NotFound );
    return ProxySession::RouteReject;
  }

  return ProxySession::RouteBySession;
}

void SBCRoutingHandler::OnOrphanedMessage(
  SIPMessage & message
)
{
  if( !message.IsRequest() )
    OnRecordRouteShift( message );
}

BOOL SBCRoutingHandler::RouteLocalRegistration( 
  B2BUAConnection & connection,
  SIPMessage & invite,
  const SIPMessage & registration
)
{
  /// lock the config to make sure no one can change data via http admin until we are done
  PWaitAndSignal configLock( PHTTPConfig::m_HTTPConfigMutex );

  SIPURI targetURI;

  if( !RouteLocalRegistration( targetURI, invite, registration ) )
    return FALSE;

  connection.AddRoute( targetURI );

  invite.SetSendAddress( targetURI );

  return TRUE;
}

BOOL SBCRoutingHandler::RouteLocalRegistration( 
  SIPURI & targetURI,
  SIPMessage & request,
  const SIPMessage & registration
)
{
  /// lock the config to make sure no one can change data via http admin until we are done
  PWaitAndSignal configLock( PHTTPConfig::m_HTTPConfigMutex );

  if( !ExtractTargetURI( targetURI, registration ) )
    return FALSE;

  OString callId = request.GetCallId();

  LOG_CONTEXT( LogInfo(), callId, "*** REGISTRATION FOUND *** -->> From: " << request.GetFromURI() 
    << " Target: " << request.GetMethod() << " " << targetURI );

  return TRUE;
}

BOOL SBCRoutingHandler::ExtractTargetURI(
  SIPURI & targetURI,
  const SIPMessage & registration
)
{
  if( registration.GetContactSize() == 0 )
    return FALSE;

  Contact contact;
  if( !registration.GetContactAt( contact, 0 ) )
    return FALSE;
  
  ContactURI contactURI;
  contact.GetURI( contactURI );
  targetURI = contactURI.GetURI();
    
  /// check the via if this is a NATted registration
  Via via;
  if( registration.GetViaAt( registration.GetViaSize() - 1, via ) 
    && via.IsBehindNAT()  
  )
  {
    targetURI.SetHost( via.GetReceiveAddress().AsString() );
    targetURI.SetPort( OString( via.GetRPort() ) );
  }
  
  if( registration.IsEncrypted() )
    targetURI.AddParameter( "ENCRYPTED", "TRUE" );

  return TRUE;
}

void SBCRoutingHandler::OnRecordRouteShift(
  SIPMessage & msg
)
{
  if( !msg.IsRequest() )
  {
    PINDEX routeCount = msg.GetRecordRouteSize();
    RecordRouteList oldList = msg.GetRecordRouteList();
    RecordRouteList newList;
    BOOL willReplaceRoutes = FALSE;

    for( PINDEX i = routeCount - 1; i >= 0; i-- )
    {
      RecordRoute recordRoute = msg.GetRecordRouteAt( i );
      RecordRoute newRoute;
      BOOL willReplaceEntry = FALSE;
      for( PINDEX j = 0; j < recordRoute.GetSize(); j++ )
      {
        RouteURI routeURI;
        OString routeShift;
        recordRoute.GetURI( routeURI, j );
        if( routeURI.GetURI().GetParameter( "route-shift", routeShift ) && !routeShift.IsEmpty() )
        { 
          willReplaceEntry = willReplaceRoutes = TRUE;
          RouteURI ruri;
          SIPURI suri( routeShift );
          ruri.SetURI( suri );
          ruri.SetLooseRouter( TRUE );
          newRoute.AddURI( ruri );
        }else
        {
          newRoute.AddURI( routeURI );
        }
      }
      newList.Append( newRoute );
    }

    if( willReplaceRoutes )
      msg.SetRecordRouteList( newList );
  }
}


BOOL SBCRoutingHandler::OnInternalResolveAddress(
  const OString & domain,
  OString & result
)
{
  /// lock the config to make sure no one can change data via http admin until we are done
  PWaitAndSignal configLock( PHTTPConfig::m_HTTPConfigMutex );
  SIPURI target;
  SIPURI current = domain;
  if( m_InternalDNSMap.FindRoute( current, target ) )
  {
    result = target.AsString();
    return TRUE;
  }else
    result = domain;
  return FALSE;
}

BOOL SBCRoutingHandler::OnGetRouteTableEntry(
  const PIPSocket::Address & destination,
  PIPSocket::Address & source,
  PIPSocket::Address & route,
  OString & device
)
{
#ifdef P_LINUX
#if ENABLE_GPL_LIBS
  return IPRoute::GetRoute( destination, route, device, source );
#else
  return FALSE;
#endif
#endif
  
#ifdef WIN32
  device = ""; route = 0;
  return SIPTransportTools::GetBestInterfaceAddress( destination, source );
#endif
  
  return FALSE;
}

